unit DebugForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, StdCtrls, Menus, Buttons, ExtCtrls;

const
  NO_DEBUG_ITEM = 18;

  LST_NONE = 0;
  LST_STOP = 1;
  LST_POINTER = 2;
  
type
  TfrmDebug = class(TForm)
    ImageList: TImageList;
    StatusBar1: TStatusBar;
    lblA: TLabel;
    lblX: TLabel;
    lblY: TLabel;
    lblS: TLabel;
    lblPC: TLabel;
    gbP: TGroupBox;
    edA: TEdit;
    edX: TEdit;
    edY: TEdit;
    edS: TEdit;
    edPC: TEdit;
    cbC: TCheckBox;
    cbZ: TCheckBox;
    cbI: TCheckBox;
    cbD: TCheckBox;
    cbB: TCheckBox;
    cbR: TCheckBox;
    cbV: TCheckBox;
    cbN: TCheckBox;
    Panel1: TPanel;
    spRun: TSpeedButton;
    MainMenu1: TMainMenu;
    Debug1: TMenuItem;
    spPause: TSpeedButton;
    spStep: TSpeedButton;
    DebugList: TListView;
    lblInterrupt: TLabel;
    spSkip: TSpeedButton;
    mnuView: TMenuItem;
    mnuPattern: TMenuItem;
    mnuName: TMenuItem;
    mnuPalette: TMenuItem;
    mnuWave: TMenuItem;
    procedure FormShow(Sender: TObject);
    procedure StepOver1Click(Sender: TObject);
    procedure cbCClick(Sender: TObject);
    procedure edAChange(Sender: TObject);
    procedure edXChange(Sender: TObject);
    procedure Run1Click(Sender: TObject);
    procedure Pause1Click(Sender: TObject);
    procedure Reset1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Breakpoints1Click(Sender: TObject);
    procedure DebugListDblClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure mnuPatternClick(Sender: TObject);
  private
    { Private declarations }
    LastDebugAddr : Word;
    DebugStop : boolean;
    PreviousItem : TListItem;
    BreakPointsList : TStringList;
    procedure GotoAddress(addr: word);
    procedure LoadPage(addr: word);
    procedure AlignRegister;
    procedure AlignItem(l : TListItem);
  public
    { Public declarations }
  end;

var
  frmDebug: TfrmDebug;

implementation

uses NesForm, M6502, PatternTableForm;

{$R *.DFM}

procedure TfrmDebug.FormCreate(Sender: TObject);
begin
  BreakPointsList := TStringList.Create;
  PreviousItem := nil;
end;

procedure TfrmDebug.FormDestroy(Sender: TObject);
begin
  BreakPointsList.free;
end;

procedure TfrmDebug.AlignRegister;
begin
  Tag := 1;
  edA.Text  := IntToHex(frmNES.cpu.A, 2);
  edX.Text  := IntToHex(frmNES.cpu.X, 2);
  edY.Text  := IntToHex(frmNES.cpu.Y, 2);
  edPC.Text := IntToHex(frmNES.cpu.PC, 4);
  edS.Text  := IntToHex(frmNES.cpu.S2 - $100, 4);

  cbC.Checked := frmNES.cpu.P and C_FLAG <> 0;
  cbZ.Checked := frmNES.cpu.P and Z_FLAG <> 0;
  cbI.Checked := frmNES.cpu.P and I_FLAG <> 0;
  cbD.Checked := frmNES.cpu.P and D_FLAG <> 0;
  cbB.Checked := frmNES.cpu.P and B_FLAG <> 0;
  cbR.Checked := frmNES.cpu.P and R_FLAG <> 0;
  cbV.Checked := frmNES.cpu.P and V_FLAG <> 0;
  cbN.Checked := frmNES.cpu.P and N_FLAG <> 0;
  Tag := 0;
end; { AlignRegister }

procedure TfrmDebug.AlignItem(l : TListItem);
begin
  if Assigned(PreviousItem) then
  begin
    if BreakPointsList.IndexOf(PreviousItem.Caption)=-1 then
      PreviousItem.ImageIndex := LST_NONE
    else
      PreviousItem.ImageIndex := LST_STOP;
  end;
  if Assigned(l) then
  begin
    l.Selected := true;
    l.ImageIndex := LST_POINTER;
    l.MakeVisible(false);
  end;
end; { AlignItem }

procedure TfrmDebug.GotoAddress(Addr: word);
var
  l : TListItem;
begin
  l := DebugList.FindCaption(1, IntToHex(Addr, 4), false, true, false);
  if Assigned(l) then
    AlignItem(l)
  else
    LoadPage(Addr);
end; { GotoAddress }

procedure TfrmDebug.LoadPage(addr: word);
var
  i : integer;
  l : TListItem;
  TheAddr : word;
begin
  TheAddr := addr;
  for i:=1 to NO_DEBUG_ITEM do
  begin
    l := DebugList.Items.Add;
    l.Caption := IntToHex(addr, 4);
    l.SubItems.Add(frmNES.cpu.DAsm(addr));
  end;
  LastDebugAddr := addr;
  l := DebugList.FindCaption(0, IntToHex(TheAddr, 4), false, true, false);
  AlignItem(l);
  {
  if Assigned(l) then
  begin
    l.ImageIndex := LST_POINTER;
    l.Selected := true;
    l.MakeVisible(false);
  end;
  }
end; { LoadPage }

procedure TfrmDebug.FormShow(Sender: TObject);
begin
  DebugStop := false;
  AlignRegister;
  LastDebugAddr := frmNES.cpu.PC;
  LoadPage(LastDebugAddr);
end;

procedure TfrmDebug.StepOver1Click(Sender: TObject);
var
  addr : word;
begin
  DebugStop := false;
  addr := frmNES.cpu.PC;
  PreviousItem := DebugList.FindCaption(0, IntToHex(addr, 4), false, true, false);
  frmNES.cpu.OpCode := frmNES.MemoryReadOpCode(addr);
  frmNES.cpu.Execute;
  GotoAddress(frmNES.cpu.PC);
  AlignRegister;
end;

procedure TfrmDebug.Run1Click(Sender: TObject);
var
  addr : word;
begin
  Repeat
    addr := frmNES.cpu.PC;
    PreviousItem := DebugList.FindCaption(0, IntToHex(addr, 4), false, true, false);
    frmNES.cpu.OpCode := frmNES.MemoryReadOpCode(addr);
    frmNES.cpu.Execute;
    GotoAddress(frmNES.cpu.PC);
    AlignRegister;
    Application.ProcessMessages;
    { Controllo se sul successivo c' un breakpoint }
    if BreakPointsList.IndexOfObject(DebugList.Selected)<>-1 then
      DebugStop := true;
  until DebugStop;
end;

procedure TfrmDebug.cbCClick(Sender: TObject);
begin
  if Tag=1 then Exit;
  if TCheckBox(Sender).Name = 'cbC' then
  begin
    if cbC.Checked then
      frmNES.cpu.P := frmNES.cpu.P or C_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor C_FLAG;
  end
  else if TCheckBox(Sender).Name = 'cbZ' then
  begin
    if cbZ.Checked then
      frmNES.cpu.P := frmNES.cpu.P or Z_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor Z_FLAG;
  end
  else if TCheckBox(Sender).Name = 'cbI' then
  begin
    if cbI.Checked then
      frmNES.cpu.P := frmNES.cpu.P or I_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor I_FLAG;
  end
  else if TCheckBox(Sender).Name = 'cbD' then
  begin
    if cbD.Checked then
      frmNES.cpu.P := frmNES.cpu.P or D_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor D_FLAG;
  end
  else if TCheckBox(Sender).Name = 'cbB' then
  begin
    if cbB.Checked then
      frmNES.cpu.P := frmNES.cpu.P or B_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor B_FLAG;
  end
  else if TCheckBox(Sender).Name = 'cbR' then
  begin
    if cbR.Checked then
      frmNES.cpu.P := frmNES.cpu.P or R_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor R_FLAG;
  end
  else if TCheckBox(Sender).Name = 'cbV' then
  begin
    if cbV.Checked then
      frmNES.cpu.P := frmNES.cpu.P or V_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor V_FLAG;
  end
  else if TCheckBox(Sender).Name = 'cbN' then
  begin
    if cbN.Checked then
      frmNES.cpu.P := frmNES.cpu.P or N_FLAG
    else
      frmNES.cpu.P := frmNES.cpu.P xor N_FLAG;
  end
end;

procedure TfrmDebug.edAChange(Sender: TObject);
begin
  if Tag=1 then Exit;
  frmNES.cpu.A := byte(StrToIntDef(edA.Text, 0));
end;

procedure TfrmDebug.edXChange(Sender: TObject);
begin
  if Tag=1 then Exit;
  frmNES.cpu.X := byte(StrToIntDef(edX.Text, 0));
end;

procedure TfrmDebug.Pause1Click(Sender: TObject);
begin
  DebugStop := true;
end;

procedure TfrmDebug.Reset1Click(Sender: TObject);
begin
  DebugStop := true;
  frmNES.cpu.Reset;
end;

procedure TfrmDebug.Breakpoints1Click(Sender: TObject);
var
  k : integer;
begin
  if Assigned(DebugList.Selected) then
  begin
    if BreakPointsList.Find(DebugList.Selected.Caption, k) then
    begin
      DebugList.Selected.ImageIndex := LST_NONE;
      BreakPointsList.Delete(k);
    end
    else
    begin
      DebugList.Selected.ImageIndex := LST_STOP;
      BreakPointsList.AddObject(DebugList.Selected.Caption, DebugList.Selected);
    end;
  end;
end;

procedure TfrmDebug.DebugListDblClick(Sender: TObject);
var
  t : TListItem;
begin
  t := DebugList.Selected; //GetItemAt(CurrentX, CurrentY);
  if Assigned(t) then
  begin
    frmNES.cpu.PC := Word(StrToIntDef('$' + t.Caption, frmNES.cpu.PC));
    AlignRegister;
  end;
end;

procedure TfrmDebug.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  DebugStop := true;
  //frmNes.cpu.Run;
end;

procedure TfrmDebug.mnuPatternClick(Sender: TObject);
var
  frmPatternTable : TfrmPatternTable;
begin
  Application.CreateForm(TfrmPatternTable, frmPatternTable);
  frmPatternTable.ShowModal;
  frmPatternTable.free;
end;

end.
