//**************************************************************************//
//**                                                                      **//
//**                            Mapper                                    **//
//**                                                                      **//
//**   Copyright (C) Dott. Cosimo De Michele 1998                         **//
//**         You are not allowed to distribuite this software             **//
//**         commercially. Please, notify me, if you make any             **//
//**         changes to this file.                                        **//
//**                                                                      **//
//**************************************************************************//

unit Mapper;

interface

type
  TMMCWrite = procedure (Addr : word; value : byte);
  { MAPPER ROUTINE }
  procedure VROM_BANK1(A : word; V : byte);
  procedure VROM_BANK4(A : word; V : byte);
  procedure VROM_BANK8(A : word; V : byte);
  procedure ROM_BANK8(A : word; V : byte);
  procedure ROM_BANK16(A : word; V : byte);
  procedure ROM_BANK32(A : word; V : byte);
  { NO MAPPER }
  function  MMCInit(MMCType : integer): boolean;
  procedure NONEInit;
  procedure NONEWrite(Addr : word; value : byte);
  { MMC1 }
  procedure MMC1Init;
  procedure MMC1Write(Addr : word; value : byte);

var
  MMCWrite : TMMCWrite;
  { MMC1 }
  MMC1_Reg : array[0..3] of byte;
  MMC1_sft,
  MMC1_buf : byte;

implementation

uses NESForm, SysUtils;

  procedure VROM_BANK1(A : word; V : byte);
  begin
    with frmNES do
    begin
      if VRomSize>0 then
        VPage[(A shr 10)] := VROM + (V * $400) - A
      else
        VPage[(A shr 10)] := nil; //@VRAM[V * $400] - A;
    end;
  end; { VROM_BANK1 }

  procedure VROM_BANK4(A : word; V : byte);
  begin
    with frmNES do
    begin
      if VRomSize>0 then
      begin
        VPage[(A shr 10)] := VROM + (V * $1000) - A;
        VPage[(A shr 10)+1] := VROM + (V * $1000) - A;
        VPage[(A shr 10)+2] := VROM + (V * $1000) - A;
        VPage[(A shr 10)+3] := VROM + (V * $1000) - A;
      end
      else
      begin
        VPage[(A shr 10)] := nil; //PChar(@VRAM + (V * $400) - A);
      end;
    end;
  end; { ROM_BANK4 }

  procedure VROM_BANK8(A : word; V : byte);
  begin
    with frmNES do
    begin
      if VRomSize>0 then
      begin
        VPage[(A shr 10)] := VROM + (V * $2000) - A;
        VPage[(A shr 10)+1] := VROM + (V * $2000) - A;
        VPage[(A shr 10)+2] := VROM + (V * $2000) - A;
        VPage[(A shr 10)+3] := VROM + (V * $2000) - A;
        VPage[(A shr 10)+4] := VROM + (V * $2000) - A;
        VPage[(A shr 10)+5] := VROM + (V * $2000) - A;
        VPage[(A shr 10)+6] := VROM + (V * $2000) - A;
        VPage[(A shr 10)+7] := VROM + (V * $2000) - A;
      end
      else
      begin
        VPage[(A shr 10)] := nil; //PChar(@VRAM + (V * $400) - A);
      end;
    end;
  end; { ROM_BANK8 }

  procedure ROM_BANK8(A : word; V : byte);
  begin
    frmNES.Page[(A shr 13)] := frmNES.ROM + (V * $2000) - A;
  end; { ROM_BANK8 }

  procedure ROM_BANK16(A : word; V : byte);
  begin
    with frmNES do
    begin
      Page[(A shr 13)]     := ROM + (V * $4000) - A;
      Page[(A shr 13) + 1] := ROM + (V * $4000) - A;
    end;
  end; { ROM_BANK16 }

  procedure ROM_BANK32(A : word; V : byte);
  begin
    with frmNES do
    begin
      Page[4] := ROM + (V * $8000) - A;
      Page[5] := ROM + (V * $8000) - A;
      Page[6] := ROM + (V * $8000) - A;
      Page[7] := ROM + (V * $8000) - A;
    end;
  end; { ROM_BANK32 }

  // ******************************************** //
  // **          NONE mapper                   ** //
  // ******************************************** //
  procedure NONEInit;
  var
    LastBank,
    i : integer;
  begin
    with frmNES do
    begin
      LastBank := (RomSize - 1);
      {
      Page[4] := ROM - $8000;
      Page[5] := ROM - $8000;
      Page[6] := ROM + (LastBank * $4000) - $C000;
      Page[7] := ROM + (LastBank * $4000) - $C000;
      }
      ROM_BANK16($8000, 0);
      ROM_BANK16($C000, LastBank);
      for i:= Low(VPage) to High(VPage) do
        if VROMSize>0 then
          VPage[i] := VROM
        else
          VPage[i] := @VRAM;
    end;
  end; { NONEInit }

  procedure NONEWrite(Addr : word; value : byte);
  begin
  end; { NONEWrite }

  // ******************************************** //
  // **          MMC1 mapper                   ** //
  // ******************************************** //
  procedure MMC1Init;
  var
    i : integer;
  begin
    for i:=0 to 3 do
      MMC1_Reg[i] := 0;
    MMC1_sft := 0;
    MMC1_buf := 0;
  end; { MMC1Init }

  procedure MMC1Write(Addr : word; value : byte);
  var
    n : integer;
  begin
    n := (Addr shr 13) - 4;
    if (value and $80)<>0 then
    begin
      MMC1_sft := 0;
      MMC1_buf := 0;
      MMC1_Reg[0] := MMC1_Reg[0] or $0C;
      Exit;
    end;
    MMC1_buf := MMC1_buf or ((value and 1) shl MMC1_sft);
    inc(MMC1_sft);
    if MMC1_sft=5 then
    begin
      MMC1_Reg[n] := MMC1_buf;
      value := MMC1_buf;
      MMC1_sft := 0;
      MMC1_buf := 0;
      case n of
        0 : begin
              with frmNES do
                if (value and 2)<>0 then
                  MirrorXor := $C00
                else
                begin
                  if Mirroring=(Value and 1) then
                    MirrorXor := ($400 shl 1)
                  else
                    MirrorXor := $400;
                end;
            end;
        1 : begin
              with frmNES do
                if (MMC1_Reg[0] and $10)<>0 then
                begin
                  VPage[0] := VROM + (value * $1000);
                  VPage[1] := VROM + (value * $1000);
                  VPage[2] := VROM + (value * $1000);
                  VPage[3] := VROM + (value * $1000);
                end
                else
                begin
                  VPage[0] := VROM + ((value shr 1) * $2000);
                end;
            end;
        2 : begin
              with frmNES do
                if (MMC1_Reg[0] and $10)<>0 then
                begin
                  VPage[($1000 shr 10)]     := VROM + (value * $1000) - $1000;
                  VPage[($1000 shr 10) + 1] := VROM + (value * $1000) - $1000;
                  VPage[($1000 shr 10) + 2] := VROM + (value * $1000) - $1000;
                  VPage[($1000 shr 10) + 3] := VROM + (value * $1000) - $1000;
                end
            end;
        3 : begin
              with frmNES do
                if (MMC1_Reg[0] and 8)<>0 then
                begin
                  if ((MMC1_Reg[0] and 4)<>0) then
                  begin
                    Page[($8000 shr 13)]     := ROM + (value * $4000) - $8000;
                    Page[($8000 shr 13) + 1] := ROM + (value * $4000) - $8000;
                  end
                  else
                  begin
                    Page[($C000 shr 13)]     := ROM + (value * $4000) - $C000;
                    Page[($C000 shr 13) + 1] := ROM + (value * $4000) - $C000;
                  end;
                end
                else
                begin
                  Page[4] := ROM + ((value shr 1) * $8000) - $8000;
                  Page[5] := ROM + ((value shr 1) * $8000) - $8000;
                  Page[6] := ROM + ((value shr 1) * $8000) - $8000;
                  Page[7] := ROM + ((value shr 1) * $8000) - $8000;
                end;
            end;
      end;
    end;
  end; { MMC1Write }



  function MMCInit(MMCType : integer): boolean;
  begin
    Result := true;
    case MMCType of
      0 : begin
            NONEInit;
            MMCWrite := NONEWrite;
          end;
      1 : begin
            NONEInit;
            MMC1Init;
            MMCWrite := MMC1Write;
          end;
    else
      frmNES.Terminal.Lines.Add('* iNES mapper #' + IntToStr(MMCType) + ' not supported.');
      Result := false;
    end;
  end; { MMCInit }

end.

